/*************************************************
* X.509 CRL Source File                          *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/x509_crl.h>
#include <botan/bigint.h>
#include <botan/oids.h>
#include <botan/pipe.h>
#include <botan/pem.h>
#include <botan/conf.h>

namespace Botan {

/*************************************************
* Load a X.509 CRL                               *
*************************************************/
X509_CRL::X509_CRL(DataSource& source, X509_Encoding encoding)
   {
   version = crl_count = 0;

   try {
      if(encoding == RAW_BER)
         decode_crl(source);
      else
         {
         Pipe ber;
         ber.process_msg(pem_decode_check_label(source, "CRL"));
         decode_crl(ber);
         }
      }
   catch(Decoding_Error& e)
      {
      throw X509_CRL_Error("CRL decoding failed");
      }
   }

/*************************************************
* Decode a BER encoded X.509 CRL                 *
*************************************************/
void X509_CRL::decode_crl(DataSource& source)
   {
   BER_Decoder ber(source);
   BER_Decoder sequence = BER::get_subsequence(ber);
   tbs_crl_data = BER::get_subsequence(sequence).get_remaining();

   AlgorithmIdentifier sig_algo_outer;
   BER::decode(sequence, sig_algo_outer);
   BER::decode(sequence, sig, BIT_STRING);
   sequence.verify_end();

   BER_Decoder tbs_crl(tbs_crl_data);

   BER::decode_optional(tbs_crl, version, INTEGER, UNIVERSAL);

   if(version != 0 && version != 1)
      throw X509_CRL_Error("Unknown X.509 CRL version " +
                           to_string(version+1));

   BER::decode(tbs_crl, sig_algo);

   if(sig_algo != sig_algo_outer)
      throw X509_CRL_Error("Algorithm identifier mismatch");

   BER::decode(tbs_crl, issuer);
   BER::decode(tbs_crl, start);
   BER::decode(tbs_crl, end);

   BER_Object next_object = tbs_crl.get_next_object();

   if(next_object.tag == ASN1_Tag(SEQUENCE | CONSTRUCTED))
      {
      BER_Decoder cert_list(next_object.value);
      while(cert_list.more_items())
         {
         BER_Decoder revoked_cert = BER::get_subsequence(cert_list);
         BigInt serial_number;
         X509_CRL::CRL_Info revoked_info;

         BER::decode(revoked_cert, serial_number);
         revoked_info.serial = encode(serial_number);
         BER::decode(revoked_cert, revoked_info.time);

         if(revoked_cert.more_items())
            {
            BER_Decoder crl_entry_exts = BER::get_subsequence(revoked_cert);
            while(crl_entry_exts.more_items())
               {
               Extension extn;
               BER::decode(crl_entry_exts, extn);
               handle_crl_entry_extension(revoked_info, extn);
               }
            }
         revoked.push_back(revoked_info);
         }
      next_object = tbs_crl.get_next_object();
      }

   if(next_object.tag == ASN1_Tag(0 | CONSTRUCTED | CONTEXT_SPECIFIC))
      {
      BER_Decoder crl_options(next_object.value);
      BER_Decoder sequence = BER::get_subsequence(crl_options);

      while(sequence.more_items())
         {
         Extension extn;
         BER::decode(sequence, extn);
         handle_crl_extension(extn);
         }
      }
   else if(next_object.tag != NO_OBJECT)
      throw X509_CRL_Error("Unknown tag in CRL");

   tbs_crl.verify_end();
   }

/*************************************************
* Decode a CRL entry extension                   *
*************************************************/
void X509_CRL::handle_crl_entry_extension(X509_CRL::CRL_Info& revoked_info,
                                          const Extension& extn)
   {
   BER_Decoder value(extn.value);

   if(extn.oid == OIDS::lookup("X509v3.ReasonCode"))
      {
      u32bit reason_code;
      BER::decode(value, reason_code, ENUMERATED, UNIVERSAL);
      if(reason_code > 8)
         reason_code = 0;
      revoked_info.reason = CRL_Code(reason_code);
      }
   else
      {
      if(extn.critical)
         {
         std::string action = Config::get_string("x509/crl/unknown_critical");
         if(action == "throw")
            throw X509_CRL_Error("Unknown critical CRL entry extension " +
                                 extn.oid.as_string());
         else if(action != "ignore")
            throw Invalid_Argument("Bad value of x509/crl/unknown_critical: "
                                   + action);
         }
      return;
      }

   value.verify_end();
   }

/*************************************************
* Decode a CRL extension                         *
*************************************************/
void X509_CRL::handle_crl_extension(const Extension& extn)
   {
   BER_Decoder value(extn.value);

   if(extn.oid == OIDS::lookup("X509v3.AuthorityKeyIdentifier"))
      {
      BER_Decoder key_id = BER::get_subsequence(value);
      BER::decode_optional_string(key_id, issuer_key_id, OCTET_STRING,
                                  ASN1_Tag(0), CONTEXT_SPECIFIC);
      }
   else
      {
      if(extn.critical)
         {
         std::string action = Config::get_string("x509/crl/unknown_critical");
         if(action == "throw")
            throw X509_CRL_Error("Unknown critical CRL extension " +
                                 extn.oid.as_string());
         else if(action != "ignore")
            throw Invalid_Argument("Bad value of x509/crl/unknown_critical: "
                                   + action);
         }
      return;
      }

   value.verify_end();
   }

/*************************************************
* Return the list of revoked certificates        *
*************************************************/
std::vector<X509_CRL::CRL_Info> X509_CRL::get_revoked() const
   {
   return revoked;
   }

/*************************************************
* Return the distinguished name of the issuer    *
*************************************************/
X509_DN X509_CRL::issuer_dn() const
   {
   return issuer;
   }

/*************************************************
* Return the key identifier of the issuer        *
*************************************************/
SecureVector<byte> X509_CRL::authority_key_id() const
   {
   return issuer_key_id;
   }

/*************************************************
* Return the TBSCertList SEQUENCE                *
*************************************************/
SecureVector<byte> X509_CRL::tbs_certlist() const
   {
   return DER::put_in_sequence(tbs_crl_data);
   }

/*************************************************
* Return the signature of this CRL               *
*************************************************/
SecureVector<byte> X509_CRL::signature() const
   {
   return sig;
   }

/*************************************************
* Return the algorithm used to sign this CRL     *
*************************************************/
AlgorithmIdentifier X509_CRL::signature_algorithm() const
   {
   return sig_algo;
   }

/*************************************************
* Return the CRL number of this CRL              *
*************************************************/
u32bit X509_CRL::crl_number() const
   {
   return crl_count;
   }

/*************************************************
* Return the issue data of the CRL               *
*************************************************/
X509_Time X509_CRL::this_update() const
   {
   return start;
   }

/*************************************************
* Return the date when a new CRL will be issued  *
*************************************************/
X509_Time X509_CRL::next_update() const
   {
   return end;
   }

}
